/**
 * CRC - Canvas Render Components (alpha)
 *
 * MIT License
 *
 * Copyright 2020 Ben Lesh <ben@benlesh.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
 * documentation files (the "Software"), to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */ "use strict";
Object.defineProperty(exports, "__esModule", {
    value: true
});
function _export(target, all) {
    for(var name in all)Object.defineProperty(target, name, {
        enumerable: true,
        get: all[name]
    });
}
_export(exports, {
    clearSharedState: ()=>clearSharedState,
    crc: ()=>crc,
    createElement: ()=>createElement,
    defineComp: ()=>defineComp,
    update: ()=>update,
    crcRef: ()=>crcRef,
    crcState: ()=>crcState,
    crcMemo: ()=>crcMemo,
    crcWhenChanged: ()=>crcWhenChanged,
    crcClick: ()=>crcClick,
    crcDblClick: ()=>crcDblClick,
    crcContextMenu: ()=>crcContextMenu,
    crcMouseMove: ()=>crcMouseMove,
    crcMouseOver: ()=>crcMouseOver,
    crcMouseOut: ()=>crcMouseOut,
    crcCursor: ()=>crcCursor,
    path: ()=>path,
    rect: ()=>rect,
    line: ()=>line,
    verticalLine: ()=>verticalLine,
    horizontalLine: ()=>horizontalLine,
    img: ()=>img,
    crcRectPath: ()=>crcRectPath,
    crcLinePath: ()=>crcLinePath,
    crcSvgPath: ()=>crcSvgPath,
    svgPath: ()=>svgPath,
    g: ()=>g
});
let refsStore = new WeakMap();
let roots = new WeakMap();
let renderIds = new WeakMap();
let _currentRefs = undefined;
let _currentRefsEntry = undefined;
let _currentCanvas = undefined;
let _currentIsMounting = false;
let _currentHookIndex = 0;
let _currentRenderContext = undefined;
let _unseenIds = undefined;
function clearSharedState() {
    refsStore = new WeakMap();
    roots = new WeakMap();
    renderIds = new WeakMap();
    anonElementNames = new WeakMap();
    _unseenIds = undefined;
    _currentRefs = undefined;
    _currentRefsEntry = undefined;
    _currentCanvas = undefined;
    _currentIsMounting = false;
    _currentHookIndex = 0;
    _currentRenderContext = undefined;
}
function crc(canvas, element) {
    if (!refsStore.has(canvas)) {
        const refs = new Map();
        refsStore.set(canvas, refs);
    }
    update(canvas, element);
}
function createElement(comp, props) {
    return {
        type: comp,
        props
    };
}
function defineComp(compFn) {
    return (props)=>createElement(compFn, props);
}
/**
 * Gets the root element for a canvas and renders it. This is the
 * entry point for rendering the entire tree.
 * @param canvas The canvas to render on
 */ function executeRender(canvas) {
    renderIds.delete(canvas);
    try {
        _currentCanvas = canvas;
        _currentRefs = refsStore.get(canvas);
        const rootElement = roots.get(canvas);
        const ctx = get2dContext(canvas);
        _currentRenderContext = ctx;
        _unseenIds = new Set(_currentRefs.keys());
        const id = 'root';
        _unseenIds.delete(id);
        render(rootElement, id);
        for (const id1 of _unseenIds){
            _currentRefs.delete(id1);
        }
    } finally{
        _currentCanvas = undefined;
        _currentRefs = undefined;
        _currentRenderContext = undefined;
        _unseenIds = undefined;
    }
}
function update(canvas, element) {
    if (element) {
        roots.set(canvas, element);
    }
    if (!roots.has(canvas) || !refsStore.has(canvas)) {
        throw new Error('Canvas does not have CRC mounted.');
    }
    if (!renderIds.has(canvas)) {
        renderIds.set(canvas, requestAnimationFrame(()=>executeRender(canvas)));
    }
}
let anonElementNames = new WeakMap();
let anon = 0;
function getAnonElemName(fn) {
    if (!anonElementNames.has(fn)) {
        anonElementNames.set(fn, `Anon${anon++}`);
    }
    return anonElementNames.get(fn);
}
function getElementTypeName(element) {
    return element.props.key || element.type.name || getAnonElemName(element.type);
}
function render(element, parentId) {
    try {
        const id = parentId + ':' + getElementTypeName(element);
        _unseenIds.delete(id);
        _currentIsMounting = !_currentRefs.has(id);
        if (_currentIsMounting) {
            _currentRefs.set(id, {
                id,
                element,
                refs: []
            });
        }
        _currentRefsEntry = _currentRefs.get(id);
        _currentHookIndex = 0;
        const ctx = _currentRenderContext;
        ctx.save();
        const result = element.type(element.props, ctx);
        if (result) {
            if (Array.isArray(result)) {
                for(let i = 0; i < result.length; i++){
                    const child = result[i];
                    render(child, id + ':' + i);
                }
            } else {
                render(result, id);
            }
        }
        ctx.restore();
    } finally{
        _currentIsMounting = false;
        _currentRefsEntry = undefined;
        _currentHookIndex = 0;
    }
}
function crcRef(init) {
    if (_currentIsMounting) {
        _currentRefsEntry.refs.push({
            type: 'ref',
            current: init
        });
    }
    return _currentRefsEntry.refs[_currentHookIndex++];
}
function crcState(init) {
    if (_currentIsMounting) {
        const ref = {
            type: 'state',
            value: init,
            setter
        };
        const canvas = _currentCanvas;
        function setter(newValue) {
            ref.value = newValue;
            update(canvas);
        }
        _currentRefsEntry.refs.push(ref);
    }
    const ref1 = _currentRefsEntry.refs[_currentHookIndex++];
    return [
        ref1.value,
        ref1.setter
    ];
}
function crcMemo(create, deps) {
    const ref = crcRef();
    // NOTE: sync updates
    crcWhenChanged(()=>{
        ref.current = create();
    }, deps);
    return ref.current;
}
function crcWhenChanged(callback, deps) {
    if (_currentIsMounting) {
        const ref = {
            type: 'whenChanged',
            lastDeps: undefined,
            teardown: undefined
        };
        _currentRefsEntry.refs.push(ref);
    }
    const hookIndex = _currentHookIndex++;
    const ref1 = _currentRefsEntry.refs[hookIndex];
    if (!ref1.lastDeps || !deps || !shallowArrayEquals(deps, ref1.lastDeps)) {
        ref1.teardown?.();
        ref1.lastDeps = deps;
        ref1.teardown = callback();
    }
}
function crcBasicMouseEvent({ type , path , handler , fill =false , lineWidth =0  }) {
    crcWhenChanged(()=>{
        const ac = new AbortController();
        const transform = _currentRenderContext.getTransform();
        _currentCanvas.addEventListener(type, (e)=>{
            const canvas = e.target;
            const [x, y] = getMouseCoordinates(e);
            const ctx = get2dContext(canvas);
            ctx.setTransform(transform);
            if (fill && ctx.isPointInPath(path, x, y) || lineWidth > 0 && ctx.isPointInStroke(path, x, y)) {
                handler({
                    x,
                    y,
                    originalEvent: e
                });
            }
        }, {
            signal: ac.signal
        });
        return ()=>{
            ac.abort();
        };
    }, [
        type,
        path,
        handler,
        fill,
        lineWidth
    ]);
}
function crcClick(config) {
    crcBasicMouseEvent({
        ...config,
        type: 'click'
    });
}
function crcDblClick(config) {
    crcBasicMouseEvent({
        ...config,
        type: 'dblclick'
    });
}
function crcContextMenu(config) {
    crcBasicMouseEvent({
        ...config,
        type: 'contextmenu'
    });
}
function crcMouseMove(config) {
    crcBasicMouseEvent({
        ...config,
        type: 'mousemove'
    });
}
function crcMouseOver(config) {
    crcMouseOverOrOut({
        ...config,
        type: 'mouseover'
    });
}
function crcMouseOut(config) {
    crcMouseOverOrOut({
        ...config,
        type: 'mouseout'
    });
}
function crcMouseOverOrOut({ type , path , handler , fill =false , lineWidth =0  }) {
    const overRef = crcRef(false);
    crcWhenChanged(()=>{
        const ac = new AbortController();
        const transform = _currentRenderContext.getTransform();
        _currentCanvas.addEventListener('mousemove', (e)=>{
            const canvas = e.target;
            const [x, y] = getMouseCoordinates(e);
            const ctx = get2dContext(canvas);
            ctx.setTransform(transform);
            const isOver = fill && ctx.isPointInPath(path, x, y) || lineWidth > 0 && ctx.isPointInStroke(path, x, y);
            const wasOver = overRef.current;
            if (isOver) {
                overRef.current = true;
                if (type === 'mouseover' && !wasOver) {
                    handler({
                        x,
                        y,
                        originalEvent: e
                    });
                }
            } else {
                overRef.current = false;
                if (type === 'mouseout' && wasOver) {
                    handler({
                        x,
                        y,
                        originalEvent: e
                    });
                }
            }
        }, {
            signal: ac.signal
        });
        _currentCanvas.addEventListener('mouseout', (e)=>{
            overRef.current = false;
            if (type === 'mouseout') {
                const [x, y] = getMouseCoordinates(e);
                handler({
                    x,
                    y,
                    originalEvent: e
                });
            }
        }, {
            signal: ac.signal
        });
        return ()=>ac.abort();
    }, [
        type,
        path,
        handler,
        fill,
        lineWidth
    ]);
}
function crcCursor({ path , style , fill =false , lineWidth =0  }) {
    crcMouseOver({
        path,
        handler: (e)=>{
            const canvas = e.originalEvent.target;
            canvas.style.cursor = style;
        },
        fill,
        lineWidth
    });
    crcMouseOut({
        path,
        handler: (e)=>{
            const canvas = e.originalEvent.target;
            canvas.style.removeProperty('cursor');
        },
        fill,
        lineWidth
    });
}
function getMouseCoordinates(event) {
    const target = event.target;
    const bounds = target.getBoundingClientRect();
    const x = event.clientX - bounds.left;
    const y = event.clientY - bounds.top;
    return [
        x,
        y
    ];
}
function shallowArrayEquals(arr1, arr2) {
    if (arr1.length !== arr2.length) {
        return false;
    }
    for(let i = 0; i < arr1.length; i++){
        if (arr1[i] !== arr2[i]) {
            return false;
        }
    }
    return true;
}
function get2dContext(canvas) {
    const ctx = canvas.getContext('2d');
    if (!ctx) {
        throw new Error('Unable to get 2d context!');
    }
    return ctx;
}
function Path(props, ctx) {
    const { alpha , path , fillStyle , strokeStyle , lineWidth  } = props;
    if (alpha) {
        ctx.globalAlpha = ctx.globalAlpha * alpha;
    }
    if (fillStyle) {
        ctx.fillStyle = fillStyle;
        ctx.fill(path);
    }
    if (strokeStyle && lineWidth && lineWidth > 0) {
        ctx.strokeStyle = strokeStyle, ctx.lineWidth = lineWidth;
        ctx.stroke(path);
    }
    const { onClick , onContextMenu , onDblClick , onMouseMove , onMouseOut , onMouseOver  } = props;
    if (onClick) {
        crcClick({
            path,
            handler: onClick,
            fill: !!fillStyle,
            lineWidth
        });
    }
    if (onContextMenu) {
        crcContextMenu({
            path,
            handler: onContextMenu,
            fill: !!fillStyle,
            lineWidth
        });
    }
    if (onDblClick) {
        crcDblClick({
            path,
            handler: onDblClick,
            fill: !!fillStyle,
            lineWidth
        });
    }
    if (onMouseOver) {
        crcMouseOver({
            path,
            handler: onMouseOver,
            fill: !!fillStyle,
            lineWidth
        });
    }
    if (onMouseMove) {
        crcMouseMove({
            path,
            handler: onMouseMove,
            fill: !!fillStyle,
            lineWidth
        });
    }
    if (onMouseOut) {
        crcMouseOut({
            path,
            handler: onMouseOut,
            fill: !!fillStyle,
            lineWidth
        });
    }
    const { cursor  } = props;
    if (cursor) {
        crcCursor({
            path,
            style: cursor,
            fill: !!fillStyle,
            lineWidth
        });
    }
}
const path = defineComp(Path);
function Rect(props, ctx) {
    const { x , y , width , height , ...pathProps } = props;
    const rectPath = crcRectPath(x, y, width, height);
    return Path({
        ...pathProps,
        path: rectPath
    }, ctx);
}
const rect = defineComp(Rect);
function Line(props, ctx) {
    const { coords , ...pathProps } = props;
    const linePath = crcLinePath(coords);
    return Path({
        ...pathProps,
        path: linePath
    }, ctx);
}
const line = defineComp(Line);
function VerticalLine(props, ctx) {
    const { x: initialX , top =0 , bottom =ctx.canvas.height , alignToPixelGrid =false , ...lineProps } = props;
    const x = alignToPixelGrid ? Math.round(initialX) - (props.lineWidth ?? 0) % 1 : initialX;
    const coords = [
        [
            x,
            top
        ],
        [
            x,
            bottom
        ]
    ];
    return Line({
        coords,
        ...lineProps
    }, ctx);
}
const verticalLine = defineComp(VerticalLine);
function HorizontalLine(props, ctx) {
    const { y: initialY , left =0 , right =ctx.canvas.width , alignToPixelGrid =false , ...lineProps } = props;
    const y = alignToPixelGrid ? Math.round(initialY) - (props.lineWidth ?? 0) % 1 : initialY;
    const coords = [
        [
            left,
            y
        ],
        [
            right,
            y
        ]
    ];
    return Line({
        coords,
        ...lineProps
    }, ctx);
}
const horizontalLine = defineComp(HorizontalLine);
function Img(props, ctx) {
    const { alpha , image , x , y , width =image.width , height =image.height , ...otherProps } = props;
    if (alpha) {
        ctx.globalAlpha = ctx.globalAlpha * alpha;
    }
    ctx.drawImage(image, x, y, width, height);
    const imagePath = crcRectPath(x, y, width, height);
    // For event handling.
    return Path({
        path: imagePath,
        ...otherProps
    }, ctx);
}
const img = defineComp(Img);
function crcRectPath(x, y, width, height) {
    return crcMemo(()=>{
        const path = new Path2D();
        path.rect(x, y, width, height);
        return path;
    }, [
        x,
        y,
        width,
        height
    ]);
}
function crcLinePath(coords) {
    return crcMemo(()=>{
        const path = new Path2D();
        for(let i = 0; i < coords.length; i++){
            const [x, y] = coords[i];
            if (i === 0) {
                path.moveTo(x, y);
            } else {
                path.lineTo(x, y);
            }
        }
        return path;
    }, coords.flat());
}
function crcSvgPath(svgPathData) {
    return crcMemo(()=>new Path2D(svgPathData), [
        svgPathData
    ]);
}
function SvgPath(props, ctx) {
    const { d , ...pathProps } = props;
    const svgPath = crcSvgPath(d);
    return Path({
        ...pathProps,
        path: svgPath
    }, ctx);
}
const svgPath = defineComp(SvgPath);
function G(props, ctx) {
    const transform = new DOMMatrix();
    const { scaleX =1 , scaleY =1  } = props;
    if (scaleX !== 1 || scaleY !== 1) {
        transform.scaleSelf(scaleX, scaleY);
    }
    const { rotate  } = props;
    if (rotate) {
        transform.rotateSelf(rotate);
    }
    const { x =0 , y =0  } = props;
    if (x || y) {
        transform.translateSelf(x, y);
    }
    const { skewX  } = props;
    if (skewX) {
        transform.skewXSelf(skewX);
    }
    const { skewY  } = props;
    if (skewY) {
        transform.skewYSelf(skewY);
    }
    ctx.setTransform(transform);
    return props.children;
}
const g = defineComp(G);

//# sourceMappingURL=index.js.map